<?php

namespace MediaWiki\Rest\PathTemplateMatcher;

use RuntimeException;

/**
 * Exception indicating incorrect REST module configuration.
 * @since 1.43
 */
class ModuleConfigurationException extends RuntimeException {
}
