.ext-WikiEditor-realtimepreview-button {
	&:hover {
		background-color: unset;
	}

	&.oo-ui-toggleWidget-on .oo-ui-labelElement-label {
		color: unset;
	}

	&.oo-ui-buttonElement-frameless.oo-ui-iconElement > .oo-ui-buttonElement-button {
		padding: 0 3px 0 0;
	}
}
