@import 'mediawiki.skin.variables.less';

.ext-WikiEditor-realtimepreview-ErrorLayout {
	display: flex;
	align-items: center;
	flex-direction: column;
	justify-content: center;
	background-color: @background-color-neutral-subtle;
	height: 100%;
	width: 100%;
	text-align: center;

	& > * {
		max-width: 268px;
		margin: 6px;
	}

	.oo-ui-buttonWidget {
		margin-top: 18px;
	}
}

.ext-WikiEditor-image-realtimepreview-error {
	height: 103px;
	width: 92px;
}
