-- This file is automatically generated using maintenance/generateSchemaChangeSql.php.
-- Source: schema/abstractSchemaChanges/patch-cu_log-add-cul_result_id.json
-- Do not modify this file directly.
-- See https://www.mediawiki.org/wiki/Manual:Schema_changes
ALTER TABLE cu_log
  ADD cul_result_id BIGINT DEFAULT 0 NOT NULL;

ALTER TABLE cu_log
  ADD cul_result_plaintext_id BIGINT DEFAULT 0 NOT NULL;
