<div class="wikieditor-toolbar-dialog-wrapper">
<fieldset><div class="wikieditor-toolbar-table-form">
	<div class="wikieditor-toolbar-field-wrapper">
		<input type="checkbox" id="wikieditor-toolbar-table-dimensions-header" checked/>
		<label for="wikieditor-toolbar-table-dimensions-header"
			rel="wikieditor-toolbar-tool-table-dimensions-header"></label>
	</div>
	<div class="wikieditor-toolbar-field-wrapper">
		<input type="checkbox" id="wikieditor-toolbar-table-wikitable" checked/>
		<label for="wikieditor-toolbar-table-wikitable" rel="wikieditor-toolbar-tool-table-wikitable"></label>
	</div>
	<div class="wikieditor-toolbar-field-wrapper">
		<input type="checkbox" id="wikieditor-toolbar-table-sortable"/>
		<label for="wikieditor-toolbar-table-sortable" rel="wikieditor-toolbar-tool-table-sortable"></label>
	</div>
	<div class="wikieditor-toolbar-table-dimension-fields">
		<div class="wikieditor-toolbar-field-wrapper">
			<label for="wikieditor-toolbar-table-dimensions-rows"
				rel="wikieditor-toolbar-tool-table-dimensions-rows"></label><br/>
			<input type="number" min="1" max="1000" id="wikieditor-toolbar-table-dimensions-rows" size="4"/>
		</div>
		<div class="wikieditor-toolbar-field-wrapper">
			<label for="wikieditor-toolbar-table-dimensions-columns"
				rel="wikieditor-toolbar-tool-table-dimensions-columns"></label><br/>
			<input type="number" min="1" max="1000" id="wikieditor-toolbar-table-dimensions-columns" size="4"/>
		</div>
	</div>
</div></fieldset>
<div class="wikieditor-toolbar-table-preview-wrapper">
	<span rel="wikieditor-toolbar-tool-table-example"></span>
	<div class="wikieditor-toolbar-table-preview-content">
		<table id="wikieditor-toolbar-table-preview" class="wikieditor-toolbar-table-preview wikitable">
		<caption rel="wikieditor-toolbar-tool-table-example-caption"></caption>
		<thead>
			<tr class="wikieditor-toolbar-table-preview-header">
				<th rel="wikieditor-toolbar-tool-table-example-header"></th>
				<th rel="wikieditor-toolbar-tool-table-example-header"></th>
				<th rel="wikieditor-toolbar-tool-table-example-header"></th>
			</tr>
		</thead><tbody>
			<tr class="wikieditor-toolbar-table-preview-hidden" style="display: none;">
			</tr><tr>
				<td rel="wikieditor-toolbar-tool-table-example-cell-text"></td>
				<td rel="wikieditor-toolbar-tool-table-example-cell-text"></td>
				<td rel="wikieditor-toolbar-tool-table-example-cell-text"></td>
			</tr><tr>
				<td rel="wikieditor-toolbar-tool-table-example-cell-text"></td>
				<td rel="wikieditor-toolbar-tool-table-example-cell-text"></td>
				<td rel="wikieditor-toolbar-tool-table-example-cell-text"></td>
			</tr><tr>
				<td rel="wikieditor-toolbar-tool-table-example-cell-text"></td>
				<td rel="wikieditor-toolbar-tool-table-example-cell-text"></td>
				<td rel="wikieditor-toolbar-tool-table-example-cell-text"></td>
			</tr>
		</tbody>
		</table>
	</div>
</div></div>
