<?php
/**
 * @license GPL-2.0-or-later
 * @file
 */

/**
 * Same as Packed except different CSS is applied to make the
 * caption only show up on hover. If a touch screen is detected,
 * falls back to PackedHoverGallery. Degrades gracefully for
 * screen readers.
 */
class PackedHoverImageGallery extends PackedOverlayImageGallery {
}
