/* Match sup style in Minerva's skins.minerva.base.styles/content/text.less */
span[ rel='mw:referencedBy' ] > a::before {
	font-size: 0.75em;
	line-height: 1; // redundant but matching declarations in minerva codebase
}
